@@ -22,6 +22,7 @@ from __future__ import absolute_import
 from .node import nullrev
 from . import (
     dagop,
+    pycompat,
     smartset,
     util,
 )
@@ -48,9 +49,6 @@ def dagwalker(repo, revs):
     returned.
 
     """
-    if not revs:
-        return
-
     gpcache = {}
 
     for rev in revs:
@@ -283,7 +281,7 @@ def _getpaddingline(echars, idx, ncols, edges):
         line.extend(echars[-(remainder * 2):])
     return line
 
-def _drawendinglines(lines, extra, edgemap, seen):
+def _drawendinglines(lines, extra, edgemap, seen, state):
     """Draw ending lines for missing parent edges
 
     None indicates an edge that ends at between this node and the next
@@ -300,7 +298,8 @@ def _drawendinglines(lines, extra, edgemap, seen):
     while edgechars and edgechars[-1] is None:
         edgechars.pop()
     shift_size = max((edgechars.count(None) * 2) - 1, 0)
-    while len(lines) < 3 + shift_size:
+    minlines = 3 if not state['graphshorten'] else 2
+    while len(lines) < minlines + shift_size:
         lines.append(extra[:])
 
     if shift_size:
@@ -321,7 +320,7 @@ def _drawendinglines(lines, extra, edgemap, seen):
                 positions[i] = max(pos, targets[i])
                 line[pos] = '/' if pos > targets[i] else extra[toshift[i]]
 
-    map = {1: '|', 2: '~'}
+    map = {1: '|', 2: '~'} if not state['graphshorten'] else {1: '~'}
     for i, line in enumerate(lines):
         if None not in line:
             continue
@@ -344,6 +343,22 @@ def asciistate():
         'graphshorten': False,
     }
 
+def outputgraph(ui, graph):
+    """outputs an ASCII graph of a DAG
+
+    this is a helper function for 'ascii' below.
+
+    takes the following arguments:
+
+    - ui to write to
+    - graph data: list of { graph nodes/edges, text }
+
+    this function can be monkey-patched by extensions to alter graph display
+    without needing to mimic all of the edge-fixup logic in ascii()
+    """
+    for (ln, logstr) in graph:
+        ui.write((ln + logstr).rstrip() + "\n")
+
 def ascii(ui, state, type, char, text, coldata):
     """prints an ASCII graph of the DAG
 
@@ -413,16 +428,16 @@ def ascii(ui, state, type, char, text, coldata):
     # shift_interline is the line containing the non-vertical
     # edges between this entry and the next
     shift_interline = echars[:idx * 2]
-    for i in xrange(2 + coldiff):
+    for i in pycompat.xrange(2 + coldiff):
         shift_interline.append(' ')
     count = ncols - idx - 1
     if coldiff == -1:
-        for i in xrange(count):
+        for i in pycompat.xrange(count):
             shift_interline.extend(['/', ' '])
     elif coldiff == 0:
         shift_interline.extend(echars[(idx + 1) * 2:ncols * 2])
     else:
-        for i in xrange(count):
+        for i in pycompat.xrange(count):
             shift_interline.extend(['\\', ' '])
 
     # draw edges from the current node to its parents
@@ -449,7 +464,7 @@ def ascii(ui, state, type, char, text, coldata):
         while len(lines) < len(text):
             lines.append(extra_interline[:])
 
-    _drawendinglines(lines, extra_interline, edgemap, seen)
+    _drawendinglines(lines, extra_interline, edgemap, seen, state)
 
     while len(text) < len(lines):
         text.append("")
@@ -457,7 +472,7 @@ def ascii(ui, state, type, char, text, coldata):
     if any(len(char) > 1 for char in edgemap.values()):
         # limit drawing an edge to the first or last N lines of the current
         # section the rest of the edge is drawn like a parent line.
-        parent = state['styles'][PARENT][-1]
+        parent = state['styles'][PARENT][-1:]
         def _drawgp(char, i):
             # should a grandparent character be drawn for this line?
             if len(char) < 2:
@@ -466,15 +481,14 @@ def ascii(ui, state, type, char, text, coldata):
             # either skip first num lines or take last num lines, based on sign
             return -num <= i if num < 0 else (len(lines) - i) <= num
         for i, line in enumerate(lines):
-            line[:] = [c[-1] if _drawgp(c, i) else parent for c in line]
+            line[:] = [c[-1:] if _drawgp(c, i) else parent for c in line]
         edgemap.update(
             (e, (c if len(c) < 2 else parent)) for e, c in edgemap.items())
 
     # print lines
     indentation_level = max(ncols, ncols + coldiff)
-    for (line, logstr) in zip(lines, text):
-        ln = "%-*s %s" % (2 * indentation_level, "".join(line), logstr)
-        ui.write(ln.rstrip() + '\n')
+    lines = ["%-*s " % (2 * indentation_level, "".join(line)) for line in lines]
+    outputgraph(ui, zip(lines, text))
 
     # ... and start over
     state['lastcoldiff'] = coldiff
